from itertools import chain


class Node:
    def __init__(self, name, default_attrs, *nodes, **attributes):
        self.default_attrs = default_attrs
        self.nodes = list(nodes)
        real_attrs = default_attrs.copy()
        real_attrs.update(attributes)
        self.attributes = real_attrs
        self.name = name

    def __repr__(self):
        attrs = self.attributes.copy()
        for k, v in self.default_attrs.items():
            if k in attrs and attrs[k] == v:
                del attrs[k]
        arg_str = ", ".join(
            chain(
                (repr(node) for node in self.nodes),
                (k + "=" + repr(v) for k, v in attrs.items()),
            )
        )
        return f"odio.{self.__class__.__name__}({arg_str})"

    def __eq__(self, other):
        return (
            isinstance(other, Node)
            and self.name == other.name
            and self.nodes == other.nodes
            and self.attrs == other.attrs
        )

    def attach(self, doc, parent_elem):
        node_elem = doc.createElement(self.name)
        parent_elem.appendChild(node_elem)
        for k, v in self.attributes.items():
            try:
                i = k.index("_")
            except ValueError:
                raise Exception(
                    f"Problem with the attribute '{k}'. Attributes must have a "
                    f"namespace prefix, eg. 'text_'."
                )
            node_elem.setAttribute(k[:i] + ":" + k[i + 1 :], v)
        if isinstance(self, H):
            node_elem.setAttribute("text:outline-level", self.name[-1])
        for node in self.nodes:
            if isinstance(node, str):
                node_elem.appendChild(doc.createTextNode(node))
            else:
                node.attach(doc, node_elem)


class H(Node):
    def __init__(self, *nodes, **attrs):
        Node.__init__(self, "text:h", {"text_style_name": "Heading 1"}, *nodes, **attrs)


class P(Node):
    def __init__(self, *nodes, **attrs):
        Node.__init__(self, "text:p", {"text_style_name": "Text Body"}, *nodes, **attrs)


class Span(Node):
    def __init__(self, *nodes, **attrs):
        Node.__init__(
            self, "text:span", {"text_style_name": "Text Body"}, *nodes, **attrs
        )



import xml.dom.minidom
import zipfile
from datetime import datetime as Datetime
from decimal import Decimal
from tempfile import NamedTemporaryFile
from xml.dom import Node

import odio
from odio.common import H, P, Span


OFFICE_VALUE_TYPE = "office:value-type"


def escape(data):
    return data.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;")


def quoteattr(data):
    data = escape(data)

    if '"' in data:
        if "'" in data:
            data = '"%s"' % data.replace('"', "&quot;")
        else:
            data = "'%s'" % data
    else:
        data = '"%s"' % data
    return data


class XmlWriter:
    def __init__(self, output):
        self.indentation = 0
        self.output = output
        self._write('<?xml version="1.0" encoding="utf-8"?>\n')

    @staticmethod
    def atts_to_str(attrs):
        if len(attrs) == 0:
            return ""
        else:
            return " " + " ".join(
                f"{k}={quoteattr(v)}" for k, v in sorted(attrs.items())
            )

    def _write(self, line, indent=True):
        if indent:
            self.output.write((" " * self.indentation * 2).encode("utf8"))
        self.output.write(line.encode("utf8"))

    def start_tag(self, name, attrs):
        self._write(f"<{name}{XmlWriter.atts_to_str(attrs)}>\n")
        self.indentation += 1

    def end_tag(self, name):
        self.indentation -= 1
        self._write(f"</{name}>\n")

    def simple_tag(self, name, attrs, contents=None):
        if contents is None:
            self._write(f"<{name}{XmlWriter.atts_to_str(attrs)}/>\n")
        else:
            self._write(f"<{name}{XmlWriter.atts_to_str(attrs)}>")
            content = escape(contents)
            if name in (
                "text:a",
                "text:h",
                "text:meta",
                "text:meta-field",
                "text:p",
                "text:ruby-base",
                "text:span",
            ):
                content = "<text:line-break/>".join(content.splitlines())
            self._write(content, indent=False)
            self._write(f"</{name}>\n", indent=False)


class SpreadsheetWriter:
    def __init__(self, f, compressed):
        self.f = f
        if compressed:
            compression = zipfile.ZIP_DEFLATED
        else:
            compression = zipfile.ZIP_STORED
        self.z = zipfile.ZipFile(f, "w", compression)
        self.z.writestr("mimetype", "application/vnd.oasis.opendocument.spreadsheet")
        self.z.writestr(
            "META-INF/manifest.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest
    manifest:version="1.2"
    xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
  <manifest:file-entry
      manifest:full-path="/"
      manifest:media-type="application/vnd.oasis.opendocument.spreadsheet"/>
  <manifest:file-entry
      manifest:full-path="settings.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="content.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="meta.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="styles.xml" manifest:media-type="text/xml"/>
</manifest:manifest>
""",
        )
        self.z.writestr(
            "meta.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-meta
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.2">
  <office:meta>
      <meta:generator>Odio</meta:generator>
  </office:meta>
</office:document-meta>
""",
        )

        self.z.writestr(
            "settings.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-settings
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.2">
</office:document-settings>
""",
        )

        self.z.writestr(
            "styles.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-styles
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
    xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
    xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
    xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
    xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
    xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
    xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
    xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
    xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
    xmlns:math="http://www.w3.org/1998/Math/MathML"
    xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
    xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"
    xmlns:dom="http://www.w3.org/2001/xml-events"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    xmlns:css3t="http://www.w3.org/TR/css3-text/"
    office:version="1.2">
</office:document-styles>
""",
        )
        self.tmp = NamedTemporaryFile()
        self.writer = XmlWriter(self.tmp)
        attrs = {
            "xmlns:office": "urn:oasis:names:tc:opendocument:xmlns:office:1.0",
            "xmlns:office": "urn:oasis:names:tc:opendocument:xmlns:office:1.0",
            "xmlns:style": "urn:oasis:names:tc:opendocument:xmlns:style:1.0",
            "xmlns:text": "urn:oasis:names:tc:opendocument:xmlns:text:1.0",
            "xmlns:table": "urn:oasis:names:tc:opendocument:xmlns:table:1.0",
            "xmlns:draw": "urn:oasis:names:tc:opendocument:xmlns:drawing:1.0",
            "xmlns:fo": "urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0",
            "xmlns:xlink": "http://www.w3.org/1999/xlink",
            "xmlns:dc": "http://purl.org/dc/elements/1.1/",
            "xmlns:meta": "urn:oasis:names:tc:opendocument:xmlns:meta:1.0",
            "xmlns:number": "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0",
            "xmlns:presentation": "urn:oasis:names:tc:opendocument:xmlns"
            ":presentation:1.0",
            "xmlns:svg": "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0",
            "xmlns:chart": "urn:oasis:names:tc:opendocument:xmlns:chart:1.0",
            "xmlns:dr3d": "urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0",
            "xmlns:math": "http://www.w3.org/1998/Math/MathML",
            "xmlns:form": "urn:oasis:names:tc:opendocument:xmlns:form:1.0",
            "xmlns:script": "urn:oasis:names:tc:opendocument:xmlns:script:1.0",
            "xmlns:dom": "http://www.w3.org/2001/xml-events",
            "xmlns:xforms": "http://www.w3.org/2002/xforms",
            "xmlns:xsd": "http://www.w3.org/2001/XMLSchema",
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xmlns:of": "urn:oasis:names:tc:opendocument:xmlns:of:1.2",
            "xmlns:xhtml": "http://www.w3.org/1999/xhtml",
            "xmlns:css3t": "http://www.w3.org/TR/css3-text/",
            "office:version": "1.2",
        }
        self.writer.start_tag("office:document-content", attrs)
        self.writer.simple_tag("office:scripts", {})
        self.writer.start_tag("office:automatic-styles", {})
        self.writer.start_tag("number:date-style", {"style:name": "date"})
        self.writer.simple_tag("number:year", {"number:style": "long"})
        self.writer.simple_tag("number:text", {}, "-")
        self.writer.simple_tag("number:month", {"number:style": "long"})
        self.writer.simple_tag("number:text", {}, "-")
        self.writer.simple_tag("number:day", {"number:style": "long"})
        self.writer.simple_tag("number:text", {}, " ")
        self.writer.simple_tag("number:hours", {"number:style": "long"})
        self.writer.simple_tag("number:text", {}, ":")
        self.writer.simple_tag("number:minutes", {"number:style": "long"})
        self.writer.end_tag("number:date-style")
        self.writer.simple_tag(
            "style:style",
            {
                "style:name": "cell_date",
                "style:family": "table-cell",
                "style:parent-style-name": "Default",
                "style:data-style-name": "date",
            },
        )
        self.writer.end_tag("office:automatic-styles")
        self.writer.start_tag("office:body", {})
        self.writer.start_tag("office:spreadsheet", {})

    def append_table(self, name, rows):
        self.writer.start_tag("table:table", {"table:name": name})
        self.writer.simple_tag("table:table-column", {})
        for row in rows:
            self.writer.start_tag("table:table-row", {})
            cells = []
            for val in row:
                atts = {}
                contents = None
                if isinstance(val, Datetime):
                    atts["office:value-type"] = "date"
                    atts["office:date-value"] = val.strftime("%Y-%m-%dT%H:%M:%S")
                    atts["table:style-name"] = "cell_date"
                elif isinstance(val, str):
                    atts["office:value-type"] = "string"
                    contents = val
                    # atts['office:string-value'] = val
                elif isinstance(val, bool):
                    atts["office:value-type"] = "boolean"
                    atts["office:boolean-value"] = "true" if val else "false"
                elif isinstance(val, (float, int, Decimal)):
                    atts["office:value-type"] = "float"
                    atts["office:value"] = str(val)
                elif isinstance(val, odio.Formula):
                    atts["table:formula"] = "of:" + str(val)
                elif val is None:
                    pass
                else:
                    atts["office:value-type"] = "string"
                    atts["office:string-value"] = str(val)

                if (
                    len(cells) > 0
                    and cells[-1]["atts"] == atts
                    and cells[-1]["contents"] == contents
                ):
                    cells[-1]["count"] += 1
                else:
                    cells.append({"count": 1, "atts": atts, "contents": contents})

            for cell in cells:
                atts = cell["atts"]
                contents = cell["contents"]
                if cell["count"] > 1:
                    atts["table:number-columns-repeated"] = str(cell["count"])
                if contents is None:
                    self.writer.simple_tag("table:table-cell", atts)
                else:
                    self.writer.start_tag("table:table-cell", atts)
                    self.writer.simple_tag("text:p", {}, contents=contents)
                    self.writer.end_tag("table:table-cell")
            self.writer.end_tag("table:table-row")

        self.writer.end_tag("table:table")

    def close(self):
        self.writer.end_tag("office:spreadsheet")
        self.writer.end_tag("office:body")
        self.writer.end_tag("office:document-content")
        self.tmp.flush()
        self.tmp.seek(0)
        self.z.write(self.tmp.name, "content.xml")
        self.z.close()
        self.tmp.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class SpreadsheetReader:
    def __init__(self, spreadsheet_elem):
        self.tables = []
        for table_elem in spreadsheet_elem.getElementsByTagName("table:table"):
            self.tables.append(TableReader(table_elem))


def _get_text(node):
    txt = [""]
    for child in node.childNodes:
        if child.nodeType == Node.TEXT_NODE:
            txt.append(child.nodeValue.strip())
        else:
            txt.append(_get_text(child))
    return "".join(txt)


class TableReader:
    def __init__(self, table_elem):
        self.name = table_elem.getAttribute("table:name")
        self.rows = []
        for row_elem in table_elem.getElementsByTagName("table:table-row"):
            row = []
            self.rows.append(row)
            for cell_elem in row_elem.getElementsByTagName("table:table-cell"):
                if cell_elem.hasAttribute("table:formula"):
                    formula = cell_elem.getAttribute("table:formula")
                    eq_idx = formula.index("=")
                    val = odio.Formula(formula[eq_idx:])
                elif cell_elem.hasAttribute(OFFICE_VALUE_TYPE):
                    val_type = cell_elem.getAttribute(OFFICE_VALUE_TYPE)
                    if val_type == "date":
                        val = Datetime.strptime(
                            cell_elem.getAttribute("office:date-value"),
                            "%Y-%m-%dT%H:%M:%S",
                        )
                    elif val_type == "string":
                        if cell_elem.hasAttribute("office:string-value"):
                            val = cell_elem.getAttribute("office:string-value")
                        else:
                            val = _get_text(cell_elem)
                    elif val_type == "float":
                        val = float(cell_elem.getAttribute("office:value"))
                    elif val_type == "boolean":
                        val = cell_elem.getAttribute("office:boolean-value") == "true"
                else:
                    val = None

                if cell_elem.hasAttribute("table:number-columns-repeated"):
                    count = int(cell_elem.getAttribute("table:number-columns-repeated"))
                else:
                    count = 1

                for i in range(count):
                    row.append(val)


class TextWriter:
    def __init__(self, f):
        self.f = f
        self.z = zipfile.ZipFile(f, "w")
        self.z.writestr("mimetype", "application/vnd.oasis.opendocument.text")
        self.z.writestr(
            "META-INF/manifest.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest
    manifest:version="1.2"
    xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
  <manifest:file-entry
      manifest:full-path="/"
      manifest:media-type="application/vnd.oasis.opendocument.text"/>
  <manifest:file-entry
      manifest:full-path="settings.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="content.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="meta.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="styles.xml" manifest:media-type="text/xml"/>
</manifest:manifest>
""",
        )
        self.z.writestr(
            "meta.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-meta
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.2">
  <office:meta>
      <meta:generator>Odio</meta:generator>
  </office:meta>
</office:document-meta>
""",
        )

        self.z.writestr(
            "settings.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-settings
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.2">
</office:document-settings>
""",
        )

        self.z.writestr(
            "styles.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-styles
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
    xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
    xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
    xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
    xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
    xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
    xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
    xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
    xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
    xmlns:math="http://www.w3.org/1998/Math/MathML"
    xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
    xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"
    xmlns:dom="http://www.w3.org/2001/xml-events"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    xmlns:css3t="http://www.w3.org/TR/css3-text/"
    office:version="1.2">
</office:document-styles>
""",
        )
        self.doc = xml.dom.minidom.parseString(
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-content
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
    xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
    xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
    xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
    xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
    xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
    xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
    xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
    xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
    xmlns:math="http://www.w3.org/1998/Math/MathML"
    xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
    xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"
    xmlns:dom="http://www.w3.org/2001/xml-events"
    xmlns:xforms="http://www.w3.org/2002/xforms"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    xmlns:css3t="http://www.w3.org/TR/css3-text/"
    office:version="1.2">
  <office:scripts/>
  <office:automatic-styles>
    <number:date-style style:name="date">
      <number:year number:style="long"/>
      <number:text>-</number:text>
      <number:month number:style="long"/>
      <number:text>-</number:text>
      <number:day number:style="long"/>
      <number:text> </number:text>
      <number:hours number:style="long"/>
      <number:text>:</number:text>
      <number:minutes number:style="long"/>
    </number:date-style>
    <style:style style:name="cell_date" style:family="table-cell"
      style:parent-style-name="Default" style:data-style-name="date"/>
  </office:automatic-styles>
  <office:body>
    <office:text>
    </office:text>
  </office:body>
</office:document-content>"""
        )
        self.text_elem = self.doc.getElementsByTagName("office:text")[0]

    def append(self, *subnodes):
        for node in subnodes:
            node.attach(self.doc, self.text_elem)

    def close(self):
        self.z.writestr("content.xml", self.doc.toprettyxml(encoding="utf-8"))
        self.z.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class TextReader:
    def __init__(self, text_elem):
        self.nodes = []
        self.attributes = {}
        for node_elem in text_elem.childNodes:
            self._parse_node(self, node_elem)

    def _parse_node(self, parent_node, node_dom):
        node_type = node_dom.nodeType
        if node_type == Node.ELEMENT_NODE:
            name = node_dom.tagName
            if name == "text:p":
                node = P()
            elif name == "text:h":
                node = H()
            elif name == "text:span":
                node = Span()
            else:
                raise Exception(f"Node name {name} not recognized.")
            if node_dom.hasAttributes():
                attrs = node_dom.attributes
                for i in range(len(attrs)):
                    attr = attrs.item(i)
                    attr_name = attr.name.replace(":", "_").replace("-", "_")
                    node.attributes[attr_name] = attr.value
            for subnode_dom in node_dom.childNodes:
                self._parse_node(node, subnode_dom)
        elif node_type == Node.TEXT_NODE:
            fnode = node_dom.nodeValue
            snode = fnode.strip()
            if len(snode) == 0:
                return
            node = snode
            rnode = node.rstrip()
            if len(rnode) < len(fnode):
                node += " "
            lnode = node.lstrip()
            if len(lnode) < len(fnode):
                node = " " + node
        else:
            raise Exception(f"Node type {node_type} not recognized.")
        parent_node.nodes.append(node)



import xml.dom.minidom
import zipfile

import odio.v1_1
import odio.v1_2
from odio.common import H, P, Span


def create_spreadsheet(f, version="1.2", compressed=True):
    if version == "1.1":
        return odio.v1_1.SpreadsheetWriter(f, compressed)
    elif version == "1.2":
        return odio.v1_2.SpreadsheetWriter(f, compressed)
    else:
        raise Exception(
            f"The version '{version}' isn't recognized. The valid version strings "
            f"are '1.1' and '1.2'."
        )


def parse_spreadsheet(f):
    with zipfile.ZipFile(f, "r") as z:
        content = z.read("content.xml")
    dom = xml.dom.minidom.parseString(content)
    version = dom.documentElement.getAttribute("office:version")
    spreadsheet_elem = dom.getElementsByTagName("office:spreadsheet")[0]

    if version == "1.1":
        return odio.v1_1.SpreadsheetReader(spreadsheet_elem)
    elif version == "1.2":
        return odio.v1_2.SpreadsheetReader(spreadsheet_elem)
    else:
        raise Exception(
            "The version '{version}' isn't recognized. The valid version strings "
            "are '1.1' and '1.2'."
        )


class Formula:
    def __init__(self, formula):
        self.formula = formula

    def __repr__(self):
        return f"odio.Formula('{self.formula}')"

    def __str__(self):
        return self.formula

    def __eq__(self, other):
        return isinstance(other, Formula) and self.formula == other.formula


def create_text(f, version="1.2"):
    if version == "1.1":
        return odio.v1_1.TextWriter(f)
    elif version == "1.2":
        return odio.v1_2.TextWriter(f)
    else:
        raise Exception(
            f"The version '{version}' isn't recognized. The valid version strings "
            f"are '1.1' and '1.2'."
        )


def parse_text(f):
    with zipfile.ZipFile(f, "r") as z:
        content = z.read("content.xml")
    f.close()
    dom = xml.dom.minidom.parseString(content)
    version = dom.documentElement.getAttribute("office:version")
    text_elem = dom.getElementsByTagName("office:text")[0]

    if version == "1.1":
        return odio.v1_1.TextReader(text_elem)
    elif version == "1.2":
        return odio.v1_2.TextReader(text_elem)
    else:
        raise Exception(
            f"The version '{version}' isn't recognized. The valid version strings "
            f"are '1.1' and '1.2'."
        )


__all__ = ["H", "P", "Span"]



import os
import zipfile

from datetime import datetime as Datetime

import odio


def normalize_walk(path):
    return list(
        [os.path.relpath(pth, path), sorted(dirs), sorted(fls)]
        for pth, dirs, fls in sorted(os.walk(path))
    )


def test_writerow(tmpdir):
    TABLE_NAME = "Plan"
    ROW = ["veni, vidi, vici", 0.3, 5, Datetime(2015, 6, 30, 16, 38), None]
    fname = tmpdir.join("actual.ods")
    with open(str(fname), "wb") as f:
        with odio.create_spreadsheet(f, "1.1") as sheet:
            table = sheet.append_table(TABLE_NAME)
            table.append_row(ROW)
    actual_dir = str(tmpdir.mkdir("actual"))
    with zipfile.ZipFile(str(fname)) as z:
        z.extractall(str(actual_dir))

    desired_dir = str(os.path.join(os.path.dirname(__file__), "unpacked"))

    actual_walk = normalize_walk(actual_dir)
    desired_walk = normalize_walk(desired_dir)
    assert actual_walk == desired_walk

    for i, (actual_pth, actual_dirs, actual_fls) in enumerate(actual_walk):
        desired_pth, desired_dirs, desired_fls = desired_walk[i]
        for j, actual_fl in enumerate(actual_fls):
            desired_fl = desired_fls[j]
            with open(
                os.path.join(actual_dir, actual_pth, actual_fl)
            ) as actual_f, open(
                os.path.join(desired_dir, desired_pth, desired_fl)
            ) as desired_f:
                assert "".join(actual_f) == "".join(desired_f)

    with open(str(fname), "rb") as f:
        sheet = odio.parse_spreadsheet(f)
        table = sheet.tables[0]
        assert table.name == TABLE_NAME
        assert table.rows[0] == ROW


def test_file_not_closed(tmpdir):
    fname = tmpdir.join("test.ods")
    f = open(str(fname), "wb")
    with odio.create_spreadsheet(f, "1.1") as sheet:
        table = sheet.append_table("Keats")
        table.append_row(("Season", "of", "mists"))
    f.seek(0)
    f.close()



import datetime
import os
import zipfile
from xml.dom.minidom import parseString

import odio


def normalized_walk(path):
    return list(
        [os.path.relpath(pth, path), sorted(dirs), sorted(fls)]
        for pth, dirs, fls in sorted(os.walk(path))
    )


def test_create_parse_spreadsheet(tmpdir):
    TABLE_NAME = "Plan"
    ROW = [
        "veni, vidi, vici",
        0.3,
        5,
        5,
        odio.Formula("=B1 + C1"),
        datetime.datetime(2015, 6, 30, 16, 38),
        None,
        "Dombey & Son",
        True,
    ]
    fname = tmpdir.join("actual.ods")
    with open(str(fname), "wb") as f, odio.create_spreadsheet(f, "1.2") as sheet:
        sheet.append_table(TABLE_NAME, [ROW])
    actual_dir = str(tmpdir.mkdir("actual"))
    with zipfile.ZipFile(str(fname)) as z:
        z.extractall(actual_dir)

    desired_dir = str(os.path.join(os.path.dirname(__file__), "unpacked"))

    actual_walk = normalized_walk(actual_dir)
    desired_walk = normalized_walk(desired_dir)
    assert actual_walk == desired_walk

    for i, (actual_pth, actual_dirs, actual_fls) in enumerate(actual_walk):
        desired_pth, desired_dirs, desired_fls = desired_walk[i]
        for j, actual_fl in enumerate(actual_fls):
            desired_fl = desired_fls[j]
            with open(
                os.path.join(actual_dir, actual_pth, actual_fl)
            ) as actual_f, open(
                os.path.join(desired_dir, desired_pth, desired_fl)
            ) as desired_f:
                ac = "".join(actual_f)
                de = "".join(desired_f)
                assert ac == de

    with open(str(fname), "rb") as f:
        sheet = odio.parse_spreadsheet(f)

    table = sheet.tables[0]
    assert table.name == TABLE_NAME
    assert table.rows[0] == ROW


def test_parse_spreadsheet_cell_p():
    xml_str = """<?xml version="1.0" encoding="UTF-8"?>
<table:table
    xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
    xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    table:name="table1">
  <table:table-row>
    <table:table-cell office:value-type="string">
      <text:p>electron</text:p>
    </table:table-cell>
  </table:table-row>
</table:table>"""
    dom = parseString(xml_str)
    result = odio.v1_2.TableReader(dom.documentElement)
    assert result.rows[0][0] == "electron"


def test_file_not_closed(tmpdir):
    fname = tmpdir.join("test.ods")
    f = open(str(fname), "wb")
    with odio.create_spreadsheet(f, "1.2") as sheet:
        sheet.append_table("Keats", ("Season", "of", "mists"))
    f.seek(0)
    f.close()


def test_get_text():
    xml_str = """<?xml version="1.0" encoding="UTF-8"?>
<cell></cell>"""
    dom = parseString(xml_str)

    val = odio.v1_2._get_text(dom)
    assert val == ""



import xml.dom.minidom
import zipfile
from datetime import datetime as Datetime


OFFICE_VALUE_TYPE = "office:value-type"


class SpreadsheetWriter:
    def __init__(self, f, compressed):
        self.f = f
        if compressed:
            compression = zipfile.ZIP_DEFLATED
        else:
            compression = zipfile.ZIP_STORED
        self.z = zipfile.ZipFile(f, "w", compression)
        self.z.writestr("mimetype", "application/vnd.oasis.opendocument.spreadsheet")
        self.z.writestr(
            "META-INF/manifest.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest
    xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
  <manifest:file-entry
      manifest:full-path="/"
      manifest:media-type="application/vnd.oasis.opendocument.spreadsheet"/>
  <manifest:file-entry
      manifest:full-path="settings.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="content.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="meta.xml" manifest:media-type="text/xml"/>
  <manifest:file-entry
      manifest:full-path="styles.xml" manifest:media-type="text/xml"/>
</manifest:manifest>""",
        )
        self.z.writestr(
            "meta.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-meta
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.1">
  <office:meta>
      <meta:generator>ODFIO</meta:generator>
  </office:meta>
</office:document-meta>""",
        )

        self.z.writestr(
            "settings.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-settings
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    office:version="1.1">
</office:document-settings>""",
        )

        self.z.writestr(
            "styles.xml",
            """<?xml version="1.0" encoding="UTF-8"?>
<office:document-styles
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
    xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
    xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
    xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
    xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
    xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
    xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
    xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
    xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
    xmlns:math="http://www.w3.org/1998/Math/MathML"
    xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
    xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"
    xmlns:dom="http://www.w3.org/2001/xml-events"
    xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    xmlns:css3t="http://www.w3.org/TR/css3-text/"
    office:version="1.1">
</office:document-styles>""",
        )

        impl = xml.dom.minidom.getDOMImplementation()
        self.doc = impl.createDocument(
            "urn:oasis:names:tc:opendocument:xmlns:office:1.0",
            "office:document-content",
            None,
        )
        document_content = self.doc.documentElement
        for name, value in (
            ("office:version", "1.1"),
            ("xmlns:chart", "urn:oasis:names:tc:opendocument:xmlns:chart:1.0"),
            ("xmlns:css3t", "http://www.w3.org/TR/css3-text/"),
            ("xmlns:dc", "http://purl.org/dc/elements/1.1/"),
            ("xmlns:dom", "http://www.w3.org/2001/xml-events"),
            ("xmlns:dr3d", "urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"),
            ("xmlns:draw", "urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"),
            (
                "xmlns:fo",
                "urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:" "1.0",
            ),
            ("xmlns:form", "urn:oasis:names:tc:opendocument:xmlns:form:1.0"),
            ("xmlns:math", "http://www.w3.org/1998/Math/MathML"),
            ("xmlns:meta", "urn:oasis:names:tc:opendocument:xmlns:meta:1.0"),
            ("xmlns:number", "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"),
            ("xmlns:of", "urn:oasis:names:tc:opendocument:xmlns:of:1.2"),
            ("xmlns:office", "urn:oasis:names:tc:opendocument:xmlns:office:1.0"),
            (
                "xmlns:presentation",
                "urn:oasis:names:tc:opendocument:xmlns:presentation:1.0",
            ),
            ("xmlns:script", "urn:oasis:names:tc:opendocument:xmlns:script:1.0"),
            ("xmlns:style", "urn:oasis:names:tc:opendocument:xmlns:style:1.0"),
            ("xmlns:svg", "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"),
            ("xmlns:table", "urn:oasis:names:tc:opendocument:xmlns:table:1.0"),
            ("xmlns:text", "urn:oasis:names:tc:opendocument:xmlns:text:1.0"),
            ("xmlns:xforms", "http://www.w3.org/2002/xforms"),
            ("xmlns:xhtml", "http://www.w3.org/1999/xhtml"),
            ("xmlns:xlink", "http://www.w3.org/1999/xlink"),
            ("xmlns:xsd", "http://www.w3.org/2001/XMLSchema"),
            ("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance"),
        ):
            document_content.setAttribute(name, value)
        scripts = self.doc.createElement("office:scripts")
        document_content.appendChild(scripts)

        automatic_styles = self.doc.createElement("office:automatic-styles")
        document_content.appendChild(automatic_styles)

        date_style = self.doc.createElement("number:date-style")
        automatic_styles.appendChild(date_style)

        date_style.setAttribute("style:name", "date")

        year = self.doc.createElement("number:year")
        date_style.appendChild(year)

        year.setAttribute("number:style", "long")
        text = self.doc.createElement("number:text")
        date_style.appendChild(text)

        text_node = self.doc.createTextNode("-")
        text.appendChild(text_node)

        month = self.doc.createElement("number:month")
        date_style.appendChild(month)

        month.setAttribute("number:style", "long")
        text = self.doc.createElement("number:text")
        date_style.appendChild(text)

        text_node = self.doc.createTextNode("-")
        text.appendChild(text_node)

        day = self.doc.createElement("number:day")
        date_style.appendChild(day)

        day.setAttribute("number:style", "long")
        text = self.doc.createElement("number:text")
        date_style.appendChild(text)

        text_node = self.doc.createTextNode(" ")
        text.appendChild(text_node)

        hours = self.doc.createElement("number:hours")
        date_style.appendChild(hours)

        hours.setAttribute("number:style", "long")
        text = self.doc.createElement("number:text")
        date_style.appendChild(text)

        text_node = self.doc.createTextNode(":")
        text.appendChild(text_node)

        minutes = self.doc.createElement("number:minutes")
        date_style.appendChild(minutes)

        minutes.setAttribute("number:style", "long")

        style = self.doc.createElement("style:style")
        automatic_styles.appendChild(style)

        style.setAttribute("style:data-style-name", "date")
        style.setAttribute("style:family", "table-cell")
        style.setAttribute("style:name", "cell_date")
        style.setAttribute("style:parent-style-name", "Default")

        body = self.doc.createElement("office:body")
        document_content.appendChild(body)
        self.spreadsheet_elem = self.doc.createElement("office:spreadsheet")
        body.appendChild(self.spreadsheet_elem)

    def append_table(self, name):
        table_elem = self.spreadsheet_elem.appendChild(
            self.doc.createElement("table:table")
        )
        table_elem.setAttribute("table:name", name)
        table_elem.appendChild(self.doc.createElement("table:table-column"))
        return Table(self.doc, table_elem)

    def close(self):
        self.z.writestr("content.xml", self.doc.toprettyxml(encoding="utf-8"))
        self.z.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class Table:
    def __init__(self, doc, table_elem):
        self.doc = doc
        self.table_elem = table_elem

    def append_row(self, vals):
        row_elem = self.table_elem.appendChild(
            self.doc.createElement("table:table-row")
        )
        for val in vals:
            cell_elem = row_elem.appendChild(self.doc.createElement("table:table-cell"))
            if isinstance(val, Datetime):
                cell_elem.setAttribute(
                    "office:date-value", val.strftime("%Y-%m-%dT%H:%M:%S")
                )
                cell_elem.setAttribute("office:value-type", "date")
                cell_elem.setAttribute("table:style-name", "cell_date")
            elif isinstance(val, str):
                cell_elem.setAttribute("office:string-value", val)
                cell_elem.setAttribute("office:value-type", "string")
            elif isinstance(val, (float, int)):
                cell_elem.setAttribute("office:value", str(val))
                cell_elem.setAttribute("office:value-type", "float")
            elif val is None:
                pass
            else:
                raise Exception(f"Type of '{val}' not recognized.")


class SpreadsheetReader:
    def __init__(self, spreadsheet_elem):
        self.tables = []
        for table_elem in spreadsheet_elem.getElementsByTagName("table:table"):
            self.tables.append(TableReader(table_elem))


class TableReader:
    def __init__(self, table_elem):
        self.name = table_elem.getAttribute("table:name")
        self.rows = []
        for row_elem in table_elem.getElementsByTagName("table:table-row"):
            row = []
            self.rows.append(row)
            for cell_elem in row_elem.getElementsByTagName("table:table-cell"):
                if cell_elem.hasAttribute(OFFICE_VALUE_TYPE):
                    val_type = cell_elem.getAttribute(OFFICE_VALUE_TYPE)
                    if val_type == "date":
                        val = Datetime.strptime(
                            cell_elem.getAttribute("office:date-value"),
                            "%Y-%m-%dT%H:%M:%S",
                        )
                    elif val_type == "string":
                        val = cell_elem.getAttribute("office:string-value")
                    elif val_type == "float":
                        val = float(cell_elem.getAttribute("office:value"))
                else:
                    val = None
                row.append(val)
